@@ -9,42 +9,29 @@
 #include <config.h>
 #endif
 
-#include <unistd.h>
-#include <string.h>
-#include <stdlib.h>
-#include <glib.h>
-
-#include <fcntl.h>
-#include <sys/types.h>
-
-#include "soup-address.h"
 #include "soup-connection.h"
-#include "soup-marshal.h"
-#include "soup-message.h"
-#include "soup-message-private.h"
+#include "soup.h"
 #include "soup-message-queue.h"
-#include "soup-misc.h"
 #include "soup-misc-private.h"
-#include "soup-socket.h"
-#include "soup-uri.h"
-#include "soup-enum-types.h"
 
 typedef struct {
 	SoupSocket  *socket;
 
-	SoupAddress *remote_addr, *tunnel_addr;
-	SoupURI     *proxy_uri;
+	SoupAddress *local_addr;
+	SoupURI *remote_uri, *proxy_uri;
+	GProxyResolver *proxy_resolver;
 	GTlsDatabase *tlsdb;
-	gboolean     ssl, ssl_strict, ssl_fallback;
+	gboolean ssl, ssl_strict, ssl_fallback;
 
 	GMainContext *async_context;
 	gboolean      use_thread_context;
 
-	SoupMessageQueueItem *cur_item;
+	SoupMessage *current_msg;
 	SoupConnectionState state;
 	time_t       unused_timeout;
 	guint        io_timeout, idle_timeout;
 	GSource     *idle_timeout_src;
+	gboolean     reusable;
 } SoupConnectionPrivate;
 #define SOUP_CONNECTION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_CONNECTION, SoupConnectionPrivate))
 
@@ -61,9 +48,9 @@ static guint signals[LAST_SIGNAL] = { 0 };
 enum {
 	PROP_0,
 
-	PROP_REMOTE_ADDRESS,
-	PROP_TUNNEL_ADDRESS,
-	PROP_PROXY_URI,
+	PROP_LOCAL_ADDRESS,
+	PROP_REMOTE_URI,
+	PROP_PROXY_RESOLVER,
 	PROP_SSL,
 	PROP_SSL_CREDS,
 	PROP_SSL_STRICT,
@@ -73,18 +60,11 @@ enum {
 	PROP_TIMEOUT,
 	PROP_IDLE_TIMEOUT,
 	PROP_STATE,
-	PROP_MESSAGE,
 
 	LAST_PROP
 };
 
-static void set_property (GObject *object, guint prop_id,
-			  const GValue *value, GParamSpec *pspec);
-static void get_property (GObject *object, guint prop_id,
-			  GValue *value, GParamSpec *pspec);
-
 static void stop_idle_timer (SoupConnectionPrivate *priv);
-static void clear_current_item (SoupConnection *conn);
 
 /* Number of seconds after which we close a connection that hasn't yet
  * been used.
@@ -94,42 +74,31 @@ static void clear_current_item (SoupConnection *conn);
 static void
 soup_connection_init (SoupConnection *conn)
 {
-	;
 }
 
 static void
-finalize (GObject *object)
+soup_connection_finalize (GObject *object)
 {
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (object);
 
-	if (priv->remote_addr)
-		g_object_unref (priv->remote_addr);
-	if (priv->tunnel_addr)
-		g_object_unref (priv->tunnel_addr);
-	if (priv->proxy_uri)
-		soup_uri_free (priv->proxy_uri);
-	if (priv->tlsdb)
-		g_object_unref (priv->tlsdb);
-	if (priv->async_context)
-		g_main_context_unref (priv->async_context);
+	g_clear_pointer (&priv->remote_uri, soup_uri_free);
+	g_clear_pointer (&priv->proxy_uri, soup_uri_free);
+	g_clear_object (&priv->tlsdb);
+	g_clear_object (&priv->proxy_resolver);
+	g_clear_object (&priv->local_addr);
+	g_clear_pointer (&priv->async_context, g_main_context_unref);
 
 	G_OBJECT_CLASS (soup_connection_parent_class)->finalize (object);
 }
 
 static void
-dispose (GObject *object)
+soup_connection_dispose (GObject *object)
 {
 	SoupConnection *conn = SOUP_CONNECTION (object);
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
 	stop_idle_timer (priv);
-	/* Make sure clear_current_item doesn't re-establish the timeout */
-	priv->idle_timeout = 0;
 
-	if (priv->cur_item) {
-		g_warning ("Disposing connection with cur_item set");
-		clear_current_item (conn);
-	}
 	if (priv->socket) {
 		g_warning ("Disposing connection while connected");
 		soup_connection_disconnect (conn);
@@ -138,6 +107,105 @@ dispose (GObject *object)
 	G_OBJECT_CLASS (soup_connection_parent_class)->dispose (object);
 }
 
+static void
+soup_connection_set_property (GObject *object, guint prop_id,
+			      const GValue *value, GParamSpec *pspec)
+{
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (object);
+
+	switch (prop_id) {
+	case PROP_LOCAL_ADDRESS:
+		priv->local_addr = g_value_dup_object (value);
+		break;
+	case PROP_REMOTE_URI:
+		priv->remote_uri = g_value_dup_boxed (value);
+		break;
+	case PROP_PROXY_RESOLVER:
+		priv->proxy_resolver = g_value_dup_object (value);
+		break;
+	case PROP_SSL:
+		priv->ssl = g_value_get_boolean (value);
+		break;
+	case PROP_SSL_CREDS:
+		if (priv->tlsdb)
+			g_object_unref (priv->tlsdb);
+		priv->tlsdb = g_value_dup_object (value);
+		break;
+	case PROP_SSL_STRICT:
+		priv->ssl_strict = g_value_get_boolean (value);
+		break;
+	case PROP_SSL_FALLBACK:
+		priv->ssl_fallback = g_value_get_boolean (value);
+		break;
+	case PROP_ASYNC_CONTEXT:
+		priv->async_context = g_value_get_pointer (value);
+		if (priv->async_context)
+			g_main_context_ref (priv->async_context);
+		break;
+	case PROP_USE_THREAD_CONTEXT:
+		priv->use_thread_context = g_value_get_boolean (value);
+		break;
+	case PROP_TIMEOUT:
+		priv->io_timeout = g_value_get_uint (value);
+		break;
+	case PROP_IDLE_TIMEOUT:
+		priv->idle_timeout = g_value_get_uint (value);
+		break;
+	case PROP_STATE:
+		soup_connection_set_state (SOUP_CONNECTION (object), g_value_get_uint (value));
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
+static void
+soup_connection_get_property (GObject *object, guint prop_id,
+			      GValue *value, GParamSpec *pspec)
+{
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (object);
+
+	switch (prop_id) {
+	case PROP_LOCAL_ADDRESS:
+		g_value_set_object (value, priv->local_addr);
+		break;
+	case PROP_REMOTE_URI:
+		g_value_set_boxed (value, priv->remote_uri);
+		break;
+	case PROP_SSL:
+		g_value_set_boolean (value, priv->ssl);
+		break;
+	case PROP_SSL_CREDS:
+		g_value_set_object (value, priv->tlsdb);
+		break;
+	case PROP_SSL_STRICT:
+		g_value_set_boolean (value, priv->ssl_strict);
+		break;
+	case PROP_SSL_FALLBACK:
+		g_value_set_boolean (value, priv->ssl_fallback);
+		break;
+	case PROP_ASYNC_CONTEXT:
+		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
+		break;
+	case PROP_USE_THREAD_CONTEXT:
+		g_value_set_boolean (value, priv->use_thread_context);
+		break;
+	case PROP_TIMEOUT:
+		g_value_set_uint (value, priv->io_timeout);
+		break;
+	case PROP_IDLE_TIMEOUT:
+		g_value_set_uint (value, priv->idle_timeout);
+		break;
+	case PROP_STATE:
+		g_value_set_enum (value, priv->state);
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+		break;
+	}
+}
+
 static void
 soup_connection_class_init (SoupConnectionClass *connection_class)
 {
@@ -146,10 +214,10 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
 	g_type_class_add_private (connection_class, sizeof (SoupConnectionPrivate));
 
 	/* virtual method override */
-	object_class->dispose = dispose;
-	object_class->finalize = finalize;
-	object_class->set_property = set_property;
-	object_class->get_property = get_property;
+	object_class->dispose = soup_connection_dispose;
+	object_class->finalize = soup_connection_finalize;
+	object_class->set_property = soup_connection_set_property;
+	object_class->get_property = soup_connection_get_property;
 
 	/* signals */
 	signals[EVENT] =
@@ -168,31 +236,31 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
 			      G_SIGNAL_RUN_FIRST,
 			      G_STRUCT_OFFSET (SoupConnectionClass, disconnected),
 			      NULL, NULL,
-			      _soup_marshal_NONE__NONE,
+			      NULL,
 			      G_TYPE_NONE, 0);
 
 	/* properties */
 	g_object_class_install_property (
-		object_class, PROP_REMOTE_ADDRESS,
-		g_param_spec_object (SOUP_CONNECTION_REMOTE_ADDRESS,
-				     "Remote address",
-				     "The address of the HTTP or proxy server",
+		object_class, PROP_LOCAL_ADDRESS,
+		g_param_spec_object (SOUP_CONNECTION_LOCAL_ADDRESS,
+				     "Local address",
+				     "Address of local end of socket",
 				     SOUP_TYPE_ADDRESS,
 				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (
-		object_class, PROP_TUNNEL_ADDRESS,
-		g_param_spec_object (SOUP_CONNECTION_TUNNEL_ADDRESS,
-				     "Tunnel address",
-				     "The address of the HTTPS server this tunnel connects to",
-				     SOUP_TYPE_ADDRESS,
-				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
-	g_object_class_install_property (
-		object_class, PROP_PROXY_URI,
-		g_param_spec_boxed (SOUP_CONNECTION_PROXY_URI,
-				    "Proxy URI",
-				    "URI of the HTTP proxy this connection connects to",
+		object_class, PROP_REMOTE_URI,
+		g_param_spec_boxed (SOUP_CONNECTION_REMOTE_URI,
+				    "Remote URI",
+				    "The URI of the HTTP server",
 				    SOUP_TYPE_URI,
-				    G_PARAM_READWRITE));
+				    G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	g_object_class_install_property (
+		object_class, PROP_PROXY_RESOLVER,
+		g_param_spec_object (SOUP_CONNECTION_PROXY_RESOLVER,
+				     "Proxy resolver",
+				     "GProxyResolver to use",
+				     G_TYPE_PROXY_RESOLVER,
+				     G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (
 		object_class, PROP_SSL,
 		g_param_spec_boolean (SOUP_CONNECTION_SSL,
@@ -255,138 +323,20 @@ soup_connection_class_init (SoupConnectionClass *connection_class)
 				   "Current state of connection",
 				   SOUP_TYPE_CONNECTION_STATE, SOUP_CONNECTION_NEW,
 				   G_PARAM_READWRITE));
-	g_object_class_install_property (
-		object_class, PROP_MESSAGE,
-		g_param_spec_object (SOUP_CONNECTION_MESSAGE,
-				     "Message",
-				     "Message being processed",
-				     SOUP_TYPE_MESSAGE,
-				     G_PARAM_READABLE));
-}
-
-
-SoupConnection *
-soup_connection_new (const char *propname1, ...)
-{
-	SoupConnection *conn;
-	va_list ap;
-
-	va_start (ap, propname1);
-	conn = (SoupConnection *)g_object_new_valist (SOUP_TYPE_CONNECTION,
-						      propname1, ap);
-	va_end (ap);
-
-	return conn;
 }
 
 static void
-set_property (GObject *object, guint prop_id,
-	      const GValue *value, GParamSpec *pspec)
+soup_connection_event (SoupConnection      *conn,
+		       GSocketClientEvent   event,
+		       GIOStream           *connection)
 {
-	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (object);
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	switch (prop_id) {
-	case PROP_REMOTE_ADDRESS:
-		priv->remote_addr = g_value_dup_object (value);
-		break;
-	case PROP_TUNNEL_ADDRESS:
-		priv->tunnel_addr = g_value_dup_object (value);
-		break;
-	case PROP_PROXY_URI:
-		if (priv->proxy_uri)
-			soup_uri_free (priv->proxy_uri);
-		priv->proxy_uri = g_value_dup_boxed (value);
-		break;
-	case PROP_SSL:
-		priv->ssl = g_value_get_boolean (value);
-		break;
-	case PROP_SSL_CREDS:
-		if (priv->tlsdb)
-			g_object_unref (priv->tlsdb);
-		priv->tlsdb = g_value_dup_object (value);
-		break;
-	case PROP_SSL_STRICT:
-		priv->ssl_strict = g_value_get_boolean (value);
-		break;
-	case PROP_SSL_FALLBACK:
-		priv->ssl_fallback = g_value_get_boolean (value);
-		break;
-	case PROP_ASYNC_CONTEXT:
-		priv->async_context = g_value_get_pointer (value);
-		if (priv->async_context)
-			g_main_context_ref (priv->async_context);
-		break;
-	case PROP_USE_THREAD_CONTEXT:
-		priv->use_thread_context = g_value_get_boolean (value);
-		break;
-	case PROP_TIMEOUT:
-		priv->io_timeout = g_value_get_uint (value);
-		break;
-	case PROP_IDLE_TIMEOUT:
-		priv->idle_timeout = g_value_get_uint (value);
-		break;
-	case PROP_STATE:
-		soup_connection_set_state (SOUP_CONNECTION (object), g_value_get_uint (value));
-		break;
-	default:
-		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-		break;
-	}
-}
+	if (!connection && priv->socket)
+		connection = soup_socket_get_connection (priv->socket);
 
-static void
-get_property (GObject *object, guint prop_id,
-	      GValue *value, GParamSpec *pspec)
-{
-	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (object);
-
-	switch (prop_id) {
-	case PROP_REMOTE_ADDRESS:
-		g_value_set_object (value, priv->remote_addr);
-		break;
-	case PROP_TUNNEL_ADDRESS:
-		g_value_set_object (value, priv->tunnel_addr);
-		break;
-	case PROP_PROXY_URI:
-		g_value_set_boxed (value, priv->proxy_uri);
-		break;
-	case PROP_SSL:
-		g_value_set_boolean (value, priv->ssl);
-		break;
-	case PROP_SSL_CREDS:
-		g_value_set_object (value, priv->tlsdb);
-		break;
-	case PROP_SSL_STRICT:
-		g_value_set_boolean (value, priv->ssl_strict);
-		break;
-	case PROP_SSL_FALLBACK:
-		g_value_set_boolean (value, priv->ssl_fallback);
-		break;
-	case PROP_ASYNC_CONTEXT:
-		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
-		break;
-	case PROP_USE_THREAD_CONTEXT:
-		g_value_set_boolean (value, priv->use_thread_context);
-		break;
-	case PROP_TIMEOUT:
-		g_value_set_uint (value, priv->io_timeout);
-		break;
-	case PROP_IDLE_TIMEOUT:
-		g_value_set_uint (value, priv->idle_timeout);
-		break;
-	case PROP_STATE:
-		g_value_set_enum (value, priv->state);
-		break;
-	case PROP_MESSAGE:
-		if (priv->cur_item)
-			g_value_set_object (value, priv->cur_item->msg);
-		else
-			g_value_set_object (value, NULL);
-		break;
-	default:
-		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
-		break;
-	}
+	g_signal_emit (conn, signals[EVENT], 0,
+		       event, connection);
 }
 
 static gboolean
@@ -419,98 +369,71 @@ stop_idle_timer (SoupConnectionPrivate *priv)
 }
 
 static void
-current_item_restarted (SoupMessage *msg, gpointer user_data)
+current_msg_got_body (SoupMessage *msg, gpointer user_data)
 {
 	SoupConnection *conn = user_data;
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
 	priv->unused_timeout = 0;
+
+	if (priv->proxy_uri &&
+	    msg->method == SOUP_METHOD_CONNECT &&
+	    SOUP_STATUS_IS_SUCCESSFUL (msg->status_code)) {
+		soup_connection_event (conn, G_SOCKET_CLIENT_PROXY_NEGOTIATED, NULL);
+
+		/* We're now effectively no longer proxying */
+		g_clear_pointer (&priv->proxy_uri, soup_uri_free);
+	}
+
+	priv->reusable = soup_message_is_keepalive (msg);
 }
 
 static void
-set_current_item (SoupConnection *conn, SoupMessageQueueItem *item)
+clear_current_msg (SoupConnection *conn)
+{
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
+	SoupMessage *msg;
+
+	msg = priv->current_msg;
+	priv->current_msg = NULL;
+
+	g_signal_handlers_disconnect_by_func (msg, G_CALLBACK (current_msg_got_body), conn);
+	g_object_unref (msg);
+}
+
+static void
+set_current_msg (SoupConnection *conn, SoupMessage *msg)
 {
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	g_return_if_fail (priv->cur_item == NULL);
+	g_return_if_fail (priv->state == SOUP_CONNECTION_IN_USE);
 
 	g_object_freeze_notify (G_OBJECT (conn));
 
+	if (priv->current_msg) {
+		g_return_if_fail (priv->current_msg->method == SOUP_METHOD_CONNECT);
+		clear_current_msg (conn);
+	}
+
 	stop_idle_timer (priv);
 
-	item->state = SOUP_MESSAGE_RUNNING;
-	priv->cur_item = item;
-	g_object_notify (G_OBJECT (conn), "message");
+	priv->current_msg = g_object_ref (msg);
+	priv->reusable = FALSE;
 
-	g_signal_connect (item->msg, "restarted",
-			  G_CALLBACK (current_item_restarted), conn);
+	g_signal_connect (msg, "got-body",
+			  G_CALLBACK (current_msg_got_body), conn);
 
-	if (item->msg->method == SOUP_METHOD_CONNECT) {
-		g_signal_emit (conn, signals[EVENT], 0,
-			       G_SOCKET_CLIENT_PROXY_NEGOTIATING,
-			       soup_socket_get_iostream (priv->socket));
-	} else if (priv->state == SOUP_CONNECTION_IDLE)
-		soup_connection_set_state (conn, SOUP_CONNECTION_IN_USE);
+	if (priv->proxy_uri && msg->method == SOUP_METHOD_CONNECT)
+		soup_connection_event (conn, G_SOCKET_CLIENT_PROXY_NEGOTIATING, NULL);
 
 	g_object_thaw_notify (G_OBJECT (conn));
 }
 
 static void
-clear_current_item (SoupConnection *conn)
-{
-	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
-
-	g_object_freeze_notify (G_OBJECT (conn));
-
-	priv->unused_timeout = 0;
-	start_idle_timer (conn);
-
-	if (priv->cur_item) {
-		SoupMessageQueueItem *item;
-
-		item = priv->cur_item;
-		priv->cur_item = NULL;
-		g_object_notify (G_OBJECT (conn), "message");
-
-		g_signal_handlers_disconnect_by_func (item->msg, G_CALLBACK (current_item_restarted), conn);
-
-		if (item->msg->method == SOUP_METHOD_CONNECT &&
-		    SOUP_STATUS_IS_SUCCESSFUL (item->msg->status_code)) {
-			g_signal_emit (conn, signals[EVENT], 0,
-				       G_SOCKET_CLIENT_PROXY_NEGOTIATED,
-				       soup_socket_get_iostream (priv->socket));
-
-			/* We're now effectively no longer proxying */
-			soup_uri_free (priv->proxy_uri);
-			priv->proxy_uri = NULL;
-		}
-
-		if (!soup_message_is_keepalive (item->msg))
-			soup_connection_disconnect (conn);
-	}
-
-	g_object_thaw_notify (G_OBJECT (conn));
-}
-
-static void
-soup_connection_event (SoupConnection      *conn,
-		       GSocketClientEvent   event,
-		       GIOStream           *connection)
-{
-	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (conn);
-
-	if (!connection && priv->socket)
-		connection = soup_socket_get_iostream (priv->socket);
-
-	g_signal_emit (conn, signals[EVENT], 0,
-		       event, connection);
-}
-
-static void
-proxy_socket_event (SoupSocket          *socket,
-		    GSocketClientEvent   event,
-		    GIOStream           *connection,
-		    gpointer             user_data)
+re_emit_socket_event (SoupSocket          *socket,
+		      GSocketClientEvent   event,
+		      GIOStream           *connection,
+		      gpointer             user_data)
 {
 	SoupConnection *conn = user_data;
 
@@ -519,19 +442,12 @@ proxy_socket_event (SoupSocket          *socket,
 		soup_connection_event (conn, event, connection);
 }
 
-static void
-socket_disconnected (SoupSocket *sock, gpointer conn)
-{
-	soup_connection_disconnect (conn);
-}
-
 typedef struct {
 	SoupConnection *conn;
 	SoupConnectionCallback callback;
 	gpointer callback_data;
 	GCancellable *cancellable;
 	guint event_id;
-	gboolean tls_handshake;
 } SoupConnectionAsyncConnectData;
 
 static void
@@ -543,15 +459,12 @@ socket_connect_finished (SoupSocket *socket, guint status, gpointer user_data)
 	g_signal_handler_disconnect (socket, data->event_id);
 
 	if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
-		g_signal_connect (priv->socket, "disconnected",
-				  G_CALLBACK (socket_disconnected), data->conn);
-
-		if (data->tls_handshake) {
+		if (priv->ssl && !priv->proxy_uri) {
 			soup_connection_event (data->conn,
 					       G_SOCKET_CLIENT_TLS_HANDSHAKED,
 					       NULL);
 		}
-		if (!priv->ssl || !priv->tunnel_addr) {
+		if (!priv->ssl || !priv->proxy_uri) {
 			soup_connection_event (data->conn,
 					       G_SOCKET_CLIENT_COMPLETE,
 					       NULL);
@@ -580,9 +493,16 @@ static void
 socket_connect_result (SoupSocket *sock, guint status, gpointer user_data)
 {
 	SoupConnectionAsyncConnectData *data = user_data;
+	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (data->conn);
 
-	if (SOUP_STATUS_IS_SUCCESSFUL (status) &&
-	    data->tls_handshake) {
+	if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
+		socket_connect_finished (sock, status, data);
+		return;
+	}
+
+	priv->proxy_uri = soup_socket_get_http_proxy_uri (priv->socket);
+
+	if (priv->ssl && !priv->proxy_uri) {
 		if (soup_socket_start_ssl (sock, data->cancellable)) {
 			soup_connection_event (data->conn,
 					       G_SOCKET_CLIENT_TLS_HANDSHAKING,
@@ -606,6 +526,7 @@ soup_connection_connect_async (SoupConnection *conn,
 {
 	SoupConnectionAsyncConnectData *data;
 	SoupConnectionPrivate *priv;
+	SoupAddress *remote_addr;
 
 	g_return_if_fail (SOUP_IS_CONNECTION (conn));
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
@@ -618,22 +539,28 @@ soup_connection_connect_async (SoupConnection *conn,
 	data->callback = callback;
 	data->callback_data = user_data;
 	data->cancellable = cancellable ? g_object_ref (cancellable) : NULL;
-	data->tls_handshake = (priv->ssl && !priv->tunnel_addr);
 
+	remote_addr = soup_address_new (priv->remote_uri->host,
+					priv->remote_uri->port);
 	priv->socket =
-		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, priv->remote_addr,
+		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, remote_addr,
 				 SOUP_SOCKET_SSL_CREDENTIALS, priv->tlsdb,
 				 SOUP_SOCKET_SSL_STRICT, priv->ssl_strict,
 				 SOUP_SOCKET_SSL_FALLBACK, priv->ssl_fallback,
 				 SOUP_SOCKET_ASYNC_CONTEXT, priv->async_context,
 				 SOUP_SOCKET_USE_THREAD_CONTEXT, priv->use_thread_context,
+				 SOUP_SOCKET_PROXY_RESOLVER, priv->proxy_resolver,
 				 SOUP_SOCKET_TIMEOUT, priv->io_timeout,
-				 "clean-dispose", TRUE,
+				 SOUP_SOCKET_CLEAN_DISPOSE, TRUE,
+				 SOUP_SOCKET_LOCAL_ADDRESS, priv->local_addr,
 				 NULL);
+	g_object_unref (remote_addr);
+
 	data->event_id = g_signal_connect (priv->socket, "event",
-					   G_CALLBACK (proxy_socket_event),
-					   conn);
-	soup_socket_connect_async (priv->socket, cancellable,
+					   G_CALLBACK (re_emit_socket_event),
+					   data->conn);
+
+	soup_socket_connect_async (priv->socket, data->cancellable,
 				   socket_connect_result, data);
 }
 
@@ -641,7 +568,8 @@ guint
 soup_connection_connect_sync (SoupConnection *conn, GCancellable *cancellable)
 {
 	SoupConnectionPrivate *priv;
-	guint status, event_id;
+	guint status, event_id = 0;
+	SoupAddress *remote_addr;
 
 	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), SOUP_STATUS_MALFORMED);
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
@@ -649,24 +577,30 @@ soup_connection_connect_sync (SoupConnection *conn, GCancellable *cancellable)
 
 	soup_connection_set_state (conn, SOUP_CONNECTION_CONNECTING);
 
+	remote_addr = soup_address_new (priv->remote_uri->host, priv->remote_uri->port);
 	priv->socket =
-		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, priv->remote_addr,
+		soup_socket_new (SOUP_SOCKET_REMOTE_ADDRESS, remote_addr,
+				 SOUP_SOCKET_PROXY_RESOLVER, priv->proxy_resolver,
 				 SOUP_SOCKET_SSL_CREDENTIALS, priv->tlsdb,
 				 SOUP_SOCKET_SSL_STRICT, priv->ssl_strict,
 				 SOUP_SOCKET_SSL_FALLBACK, priv->ssl_fallback,
 				 SOUP_SOCKET_FLAG_NONBLOCKING, FALSE,
 				 SOUP_SOCKET_TIMEOUT, priv->io_timeout,
-				 "clean-dispose", TRUE,
+				 SOUP_SOCKET_CLEAN_DISPOSE, TRUE,
+				 SOUP_SOCKET_LOCAL_ADDRESS, priv->local_addr,
 				 NULL);
+	g_object_unref (remote_addr);
 
 	event_id = g_signal_connect (priv->socket, "event",
-				     G_CALLBACK (proxy_socket_event), conn);
+				     G_CALLBACK (re_emit_socket_event), conn);
 	status = soup_socket_connect_sync (priv->socket, cancellable);
 
 	if (!SOUP_STATUS_IS_SUCCESSFUL (status))
 		goto fail;
-		
-	if (priv->ssl && !priv->tunnel_addr) {
+
+	priv->proxy_uri = soup_socket_get_http_proxy_uri (priv->socket);
+
+	if (priv->ssl && !priv->proxy_uri) {
 		if (!soup_socket_start_ssl (priv->socket, cancellable))
 			status = SOUP_STATUS_SSL_FAILED;
 		else {
@@ -686,10 +620,7 @@ soup_connection_connect_sync (SoupConnection *conn, GCancellable *cancellable)
 	}
 
 	if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
-		g_signal_connect (priv->socket, "disconnected",
-				  G_CALLBACK (socket_disconnected), conn);
-
-		if (!priv->ssl || !priv->tunnel_addr) {
+		if (!priv->ssl || !priv->proxy_uri) {
 			soup_connection_event (conn,
 					       G_SOCKET_CLIENT_COMPLETE,
 					       NULL);
@@ -697,16 +628,10 @@ soup_connection_connect_sync (SoupConnection *conn, GCancellable *cancellable)
 		soup_connection_set_state (conn, SOUP_CONNECTION_IN_USE);
 		priv->unused_timeout = time (NULL) + SOUP_CONNECTION_UNUSED_TIMEOUT;
 		start_idle_timer (conn);
-	} else {
-	fail:
-		if (priv->socket) {
-			soup_socket_disconnect (priv->socket);
-			g_object_unref (priv->socket);
-			priv->socket = NULL;
-		}
 	}
 
-	if (priv->socket)
+ fail:
+	if (priv->socket && event_id)
 		g_signal_handler_disconnect (priv->socket, event_id);
 
 	if (priv->proxy_uri != NULL)
@@ -714,15 +639,15 @@ soup_connection_connect_sync (SoupConnection *conn, GCancellable *cancellable)
 	return status;
 }
 
-SoupAddress *
-soup_connection_get_tunnel_addr (SoupConnection *conn)
+gboolean
+soup_connection_is_tunnelled (SoupConnection *conn)
 {
 	SoupConnectionPrivate *priv;
 
-	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), NULL);
+	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	return priv->tunnel_addr;
+	return priv->ssl && priv->proxy_uri != NULL;
 }
 
 guint
@@ -730,24 +655,22 @@ soup_connection_start_ssl_sync (SoupConnection *conn,
 				GCancellable   *cancellable)
 {
 	SoupConnectionPrivate *priv;
-	const char *server_name;
 	guint status;
 
 	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), FALSE);
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
 
-	server_name = soup_address_get_name (priv->tunnel_addr ?
-					     priv->tunnel_addr :
-					     priv->remote_addr);
-	if (!soup_socket_start_proxy_ssl (priv->socket, server_name,
+	if (!soup_socket_start_proxy_ssl (priv->socket,
+					  priv->remote_uri->host,
 					  cancellable))
 		return SOUP_STATUS_SSL_FAILED;
 
 	soup_connection_event (conn, G_SOCKET_CLIENT_TLS_HANDSHAKING, NULL);
 	status = soup_socket_handshake_sync (priv->socket, cancellable);
-	if (status == SOUP_STATUS_OK)
+	if (status == SOUP_STATUS_OK) {
 		soup_connection_event (conn, G_SOCKET_CLIENT_TLS_HANDSHAKED, NULL);
-	else if (status == SOUP_STATUS_TLS_FAILED) {
+		soup_connection_event (conn, G_SOCKET_CLIENT_COMPLETE, NULL);
+	} else if (status == SOUP_STATUS_TLS_FAILED) {
 		priv->ssl_fallback = TRUE;
 		status = SOUP_STATUS_TRY_AGAIN;
 	}
@@ -761,9 +684,10 @@ start_ssl_completed (SoupSocket *socket, guint status, gpointer user_data)
 	SoupConnectionAsyncConnectData *data = user_data;
 	SoupConnectionPrivate *priv = SOUP_CONNECTION_GET_PRIVATE (data->conn);
 
-	if (status == SOUP_STATUS_OK)
+	if (status == SOUP_STATUS_OK) {
 		soup_connection_event (data->conn, G_SOCKET_CLIENT_TLS_HANDSHAKED, NULL);
-	else if (status == SOUP_STATUS_TLS_FAILED) {
+		soup_connection_event (data->conn, G_SOCKET_CLIENT_COMPLETE, NULL);
+	} else if (status == SOUP_STATUS_TLS_FAILED) {
 		priv->ssl_fallback = TRUE;
 		status = SOUP_STATUS_TRY_AGAIN;
 	}
@@ -789,7 +713,6 @@ soup_connection_start_ssl_async (SoupConnection   *conn,
 				 gpointer          user_data)
 {
 	SoupConnectionPrivate *priv;
-	const char *server_name;
 	SoupConnectionAsyncConnectData *data;
 	GMainContext *async_context;
 
@@ -806,10 +729,8 @@ soup_connection_start_ssl_async (SoupConnection   *conn,
 	else
 		async_context = priv->async_context;
 
-	server_name = soup_address_get_name (priv->tunnel_addr ?
-					     priv->tunnel_addr :
-					     priv->remote_addr);
-	if (!soup_socket_start_proxy_ssl (priv->socket, server_name,
+	if (!soup_socket_start_proxy_ssl (priv->socket,
+					  priv->remote_uri->host,
 					  cancellable)) {
 		soup_add_completion (async_context,
 				     idle_start_ssl_completed, data);
@@ -842,11 +763,11 @@ soup_connection_disconnect (SoupConnection *conn)
 		soup_connection_set_state (conn, SOUP_CONNECTION_DISCONNECTED);
 
 	if (priv->socket) {
-		g_signal_handlers_disconnect_by_func (priv->socket,
-						      socket_disconnected, conn);
-		soup_socket_disconnect (priv->socket);
-		g_object_unref (priv->socket);
+		SoupSocket *socket = priv->socket;
+
 		priv->socket = NULL;
+		soup_socket_disconnect (socket);
+		g_object_unref (socket);
 	}
 
 	if (old_state != SOUP_CONNECTION_DISCONNECTED)
@@ -861,6 +782,14 @@ soup_connection_get_socket (SoupConnection *conn)
 	return SOUP_CONNECTION_GET_PRIVATE (conn)->socket;
 }
 
+SoupURI *
+soup_connection_get_remote_uri (SoupConnection *conn)
+{
+	g_return_val_if_fail (SOUP_IS_CONNECTION (conn), NULL);
+
+	return SOUP_CONNECTION_GET_PRIVATE (conn)->remote_uri;
+}
+
 SoupURI *
 soup_connection_get_proxy_uri (SoupConnection *conn)
 {
@@ -901,7 +830,6 @@ void
 soup_connection_set_state (SoupConnection *conn, SoupConnectionState state)
 {
 	SoupConnectionPrivate *priv;
-	SoupConnectionState old_state;
 
 	g_return_if_fail (SOUP_IS_CONNECTION (conn));
 	g_return_if_fail (state >= SOUP_CONNECTION_NEW &&
@@ -910,14 +838,25 @@ soup_connection_set_state (SoupConnection *conn, SoupConnectionState state)
 	g_object_freeze_notify (G_OBJECT (conn));
 
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
-	old_state = priv->state;
-	priv->state = state;
-	if ((state == SOUP_CONNECTION_IDLE ||
-	     state == SOUP_CONNECTION_DISCONNECTED) &&
-	    old_state == SOUP_CONNECTION_IN_USE)
-		clear_current_item (conn);
 
-	g_object_notify (G_OBJECT (conn), "state");
+	if (priv->current_msg) {
+		g_warn_if_fail (state == SOUP_CONNECTION_IDLE ||
+				state == SOUP_CONNECTION_DISCONNECTED);
+		clear_current_msg (conn);
+	}
+
+	if (state == SOUP_CONNECTION_IDLE && !priv->reusable) {
+		/* This will recursively call set_state() */
+		soup_connection_disconnect (conn);
+	} else {
+		priv->state = state;
+
+		if (priv->state == SOUP_CONNECTION_IDLE)
+			start_idle_timer (conn);
+
+		g_object_notify (G_OBJECT (conn), "state");
+	}
+
 	g_object_thaw_notify (G_OBJECT (conn));
 }
 
@@ -946,9 +885,13 @@ soup_connection_send_request (SoupConnection          *conn,
 	g_return_if_fail (SOUP_IS_CONNECTION (conn));
 	g_return_if_fail (item != NULL);
 	priv = SOUP_CONNECTION_GET_PRIVATE (conn);
-	g_return_if_fail (priv->state != SOUP_CONNECTION_NEW && priv->state != SOUP_CONNECTION_DISCONNECTED);
+	g_return_if_fail (priv->state != SOUP_CONNECTION_NEW &&
+			  priv->state != SOUP_CONNECTION_DISCONNECTED);
+
+	if (item->msg != priv->current_msg)
+		set_current_msg (conn, item->msg);
+	else
+		priv->reusable = FALSE;
 
-	if (item != priv->cur_item)
-		set_current_item (conn, item);
 	soup_message_send_request (item, completion_cb, user_data);
 }
