@@ -5,7 +5,7 @@
 
 #include "test-utils.h"
 
-#include "libsoup/soup-connection.h"
+#include <gio/gnetworking.h>
 
 SoupServer *server;
 SoupURI *base_uri;
@@ -301,7 +301,6 @@ do_timeout_test_for_session (SoupSession *session)
 static void
 do_timeout_req_test_for_session (SoupSession *session)
 {
-	SoupRequester *requester;
 	SoupRequest *req;
 	SoupMessage *msg;
 	GInputStream *stream;
@@ -310,26 +309,30 @@ do_timeout_req_test_for_session (SoupSession *session)
 	GError *error = NULL;
 	int i;
 
-	requester = soup_requester_new ();
-	soup_session_add_feature (session, SOUP_SESSION_FEATURE (requester));
-	g_object_unref (requester);
-
 	g_signal_connect (session, "request-started",
 			  G_CALLBACK (request_started_socket_collector),
 			  &sockets);
 
 	debug_printf (1, "    First request\n");
 	timeout_uri = soup_uri_new_with_base (base_uri, "/timeout-persistent");
-	req = soup_requester_request_uri (requester, timeout_uri, NULL);
+	req = soup_session_request_uri (session, timeout_uri, NULL);
 	soup_uri_free (timeout_uri);
 
-	stream = soup_test_request_send (req, NULL, &error);
+	stream = soup_test_request_send (req, NULL, 0, &error);
 	if (!stream) {
 		debug_printf (1, "      Unexpected error on send: %s\n",
 			      error->message);
 		errors++;
 		g_clear_error (&error);
 	} else {
+		soup_test_request_read_all (req, stream, NULL, &error);
+		if (error) {
+			debug_printf (1, "  Unexpected error on read: %s\n",
+				      error->message);
+			errors++;
+			g_clear_error (&error);
+		}
+
 		soup_test_request_close_stream (req, stream, NULL, &error);
 		if (error) {
 			debug_printf (1, "  Unexpected error on close: %s\n",
@@ -348,9 +351,9 @@ do_timeout_req_test_for_session (SoupSession *session)
 	g_object_unref (req);
 
 	debug_printf (1, "    Second request\n");
-	req = soup_requester_request_uri (requester, base_uri, NULL);
+	req = soup_session_request_uri (session, base_uri, NULL);
 
-	stream = soup_test_request_send (req, NULL, &error);
+	stream = soup_test_request_send (req, NULL, 0, &error);
 	if (!stream) {
 		debug_printf (1, "      Unexpected error on send: %s\n",
 			      error->message);
@@ -405,6 +408,7 @@ do_persistent_connection_timeout_test (void)
 	do_timeout_test_for_session (session);
 	soup_test_session_abort_unref (session);
 
+	debug_printf (1, "  Async session, request API\n");
 	session = soup_test_session_new (SOUP_TYPE_SESSION_ASYNC,
 					 SOUP_SESSION_USE_THREAD_CONTEXT, TRUE,
 					 NULL);
@@ -416,6 +420,7 @@ do_persistent_connection_timeout_test (void)
 	do_timeout_test_for_session (session);
 	soup_test_session_abort_unref (session);
 
+	debug_printf (1, "  Sync session, request API\n");
 	session = soup_test_session_new (SOUP_TYPE_SESSION_SYNC, NULL);
 	do_timeout_req_test_for_session (session);
 	soup_test_session_abort_unref (session);
@@ -490,6 +495,8 @@ do_max_conns_test_for_session (SoupSession *session)
 
 	msgs_done = 0;
 	g_idle_add (idle_start_server, NULL);
+	if (quit_loop_timeout)
+		g_source_remove (quit_loop_timeout);
 	quit_loop_timeout = g_timeout_add (1000, quit_loop, NULL);
 	g_main_loop_run (max_conns_loop);
 
@@ -512,8 +519,10 @@ do_max_conns_test_for_session (SoupSession *session)
 	}
 
 	g_main_loop_unref (max_conns_loop);
-	if (quit_loop_timeout)
+	if (quit_loop_timeout) {
 		g_source_remove (quit_loop_timeout);
+		quit_loop_timeout = 0;
+	}
 
 	for (i = 0; i < TEST_CONNS; i++)
 		g_object_unref (msgs[i]);
@@ -713,11 +722,10 @@ static void
 connection_state_changed (GObject *object, GParamSpec *param,
 			  gpointer user_data)
 {
-	SoupConnection *conn = SOUP_CONNECTION (object);
 	SoupConnectionState *state = user_data;
 	SoupConnectionState new_state;
 
-	new_state = soup_connection_get_state (conn);
+	g_object_get (object, "state", &new_state, NULL);
 	if (state_transitions[*state] != new_state) {
 		debug_printf (1, "      Unexpected transition: %s -> %s\n",
 			      state_names[*state], state_names[new_state]);
@@ -731,12 +739,12 @@ connection_state_changed (GObject *object, GParamSpec *param,
 }
 
 static void
-connection_created (SoupSession *session, SoupConnection *conn,
+connection_created (SoupSession *session, GObject *conn,
 		    gpointer user_data)
 {
 	SoupConnectionState *state = user_data;
 
-	*state = soup_connection_get_state (conn);
+	g_object_get (conn, "state", state, NULL);
 	if (*state != SOUP_CONNECTION_NEW) {
 		debug_printf (1, "      Unexpected initial state: %d\n",
 			      *state);
@@ -777,8 +785,11 @@ do_connection_state_test_for_session (SoupSession *session)
 	debug_printf (1, "    http\n");
 	do_one_connection_state_test (session, HTTP_SERVER);
 
-	debug_printf (1, "    https\n");
-	do_one_connection_state_test (session, HTTPS_SERVER);
+	if (tls_available) {
+		debug_printf (1, "    https\n");
+		do_one_connection_state_test (session, HTTPS_SERVER);
+	} else
+		debug_printf (1, "    https -- SKIPPING\n");
 
 	proxy_uri = soup_uri_new (HTTP_PROXY);
 	g_object_set (G_OBJECT (session),
@@ -789,8 +800,11 @@ do_connection_state_test_for_session (SoupSession *session)
 	debug_printf (1, "    http with proxy\n");
 	do_one_connection_state_test (session, HTTP_SERVER);
 
-	debug_printf (1, "    https with proxy\n");
-	do_one_connection_state_test (session, HTTPS_SERVER);
+	if (tls_available) {
+		debug_printf (1, "    https with proxy\n");
+		do_one_connection_state_test (session, HTTPS_SERVER);
+	} else
+		debug_printf (1, "    https with proxy -- SKIPPING\n");
 }
 
 static void
@@ -892,8 +906,11 @@ do_connection_event_test_for_session (SoupSession *session)
 	debug_printf (1, "    http\n");
 	do_one_connection_event_test (session, HTTP_SERVER, "rRcCx");
 
-	debug_printf (1, "    https\n");
-	do_one_connection_event_test (session, HTTPS_SERVER, "rRcCtTx");
+	if (tls_available) {
+		debug_printf (1, "    https\n");
+		do_one_connection_event_test (session, HTTPS_SERVER, "rRcCtTx");
+	} else
+		debug_printf (1, "    https -- SKIPPING\n");
 
 	proxy_uri = soup_uri_new (HTTP_PROXY);
 	g_object_set (G_OBJECT (session),
@@ -904,8 +921,11 @@ do_connection_event_test_for_session (SoupSession *session)
 	debug_printf (1, "    http with proxy\n");
 	do_one_connection_event_test (session, HTTP_SERVER, "rRcCx");
 
-	debug_printf (1, "    https with proxy\n");
-	do_one_connection_event_test (session, HTTPS_SERVER, "rRcCpPtTx");
+	if (tls_available) {
+		debug_printf (1, "    https with proxy\n");
+		do_one_connection_event_test (session, HTTPS_SERVER, "rRcCpPtTx");
+	} else
+		debug_printf (1, "    https with proxy -- SKIPPING\n");
 }
 
 static void
